/* Author: Domagoj Babic 2004.
 *
 * $Id: hsat.h,v 1.13 2005/02/21 23:44:09 root Exp $
 *
 */

/***********************************************************************
*
*  Copyright (C) 2004-2005, by Domagoj Babic (Author). All rights 
*  reserved. By using this software the USER indicates that he or she 
*  has read, understood and will comply with the following:
*
*  This software may not be called "HyperSAT" if it has been modified in 
*  any way, without the specific prior written permission of the Author.
*
*  Author hereby grants USER nonexclusive permission to use, copy and/or 
*  modify this software for internal, noncommercial, research purposes 
*  only. Any distribution, including commercial sale or license, of this 
*  software, copies of the software, its associated documentation and/or 
*  modifications of either is strictly prohibited without the prior 
*  consent from the Author. Appropriate copyright notice shall be placed 
*  on all software copies, and a complete copy of this notice shall be 
*  included in all copies of the associated documentation.
*
*  THE SOFTWARE IS PROVIDED "AS-IS" AND AUTHOR MAKES NO REPRESENTATIONS 
*  OR WARRANTIES, EXPRESS OR IMPLIED, BY WAY OF EXAMPLE, BUT NOT 
*  LIMITATION. AUTHOR MAKES NO REPRESENTATIONS OR WARRANTIES OF 
*  MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR PURPOSE OR THAT THE
*  USE OF THE SOFTWARE WILL NOT INFRINGE ANY PATENTS, COPYRIGHTS
*  TRADEMARKS OR OTHER RIGHTS. AUTHOR SHALL NOT BE LIABLE FOR ANY
*  LIABILITY OR DAMAGES WITH RESPECT TO ANY CLAIM BY LICENSEE OR ANY
*  THIRD PARTY ON ACCOUNT OF, OR ARISING FROM THE LICENSE, OR ANY
*  SUBLICENSE OR USE OF THE SOFTWARE OR ANY SERVICE OR SUPPORT.
*
*  Author shall not be liable under any circumstances for any direct, 
*  indirect, special, incidental, or consequential damages with respect 
*  to any claim by USER or any third party on account of or arising from 
*  the use, or inability to use, this software or its associated 
*  documentation, even if the Author has been advised of the possibility 
*  of those damages.
*
*  The above copyright notice must appear in all copies of the software.
*
************************************************************************
*
*  Support notice: Unfortunatelly, the author can't guarantee any 
*  support. The solver has been written exclusively for research 
*  and experimental purposes, so documentation and comments are scarce.
*  Most likely, this will not change in the near future, but feel free
*  to check my webpage occasionally, as new releases (if there will be
*  any) will be made available at my website:
*  http:/www.cs.ubc.ca/~babic/
*
***********************************************************************/

#include <ctype.h>
#include "utils.h"
#include "datas.h"
#include "fheap.h"
#include "hcube.h"

// MARK -   for conflict analysis
// ADDED -  1 if the variable has been added to conflict
//          clause
// WARRNING: VALUE must always be 1 !!                             
enum varflags_e {VALUE = 1, ASSIGNED = 2, BOTH_DIR = 4, 
    HASCUBE = 8, DEPENDANT = 16, MARK = 32, ADDED = 64,
    FIXED = 128, BRANCH_BCT_ND = 256, LITERAL_BCT_ND = 512,
    FIXED_SIGN = 1024, ELIMINATED = 2048};

enum result_e {CONFLICT = 0, NO_CONFLICT = 1};

enum cache_update_policy_e {NO_UPDATES, DEC_CONFLICTING_CLS, 
    UIP_CONFLICTING_CLS};

enum var_update_policy_e {PRI_UIP_CONSTRUCTION, 
    PRI_DEC_CONSTRUCTION, PRI_UPDATE_ALL};

enum periodic_events_e {EVNT_CONFLICT, EVNT_DECISION, 
    EVNT_DEC_FIX_LEARN, EVNT_ASSIGNMENT, EVNT_UIP_CONSTRUCT, 
    EVNT_DEC_CONSTRUCT, EVNT_IMP_CHAIN, EVNT_IMPLICATION, 
    EVNT_SCHEDULE_IMPS, EVNT_IMP_RESOLUTION, EVNT_NEW_HCUBE_BRANCHES,
    EVNT_NEW_HCUBE_NODES, EVNT_USED_HCUBE_BRANCH, 
    EVNT_USED_HCUBE_NODE, EVNT_HCUBE_INTERSECTION, EVNT_HC_INTERS_CONF,
    EVNT_RESTART};

typedef struct params_t {
    // priority handling parameters
    uint priInitial;
    uint priIncrement;
    uint priScalePeriod;
    // priority boost for literals in binary clauses
    uint priBinBoost;
    // number of literals at the end of the clauses whose
    // priorities get updated (the end because those are
    // probably the most recently assigned ones, beside
    // watched literals)
    uint priLitsUpdated;
    
    
    /* search for ultimate priority variables only when
    // the solver backtracks over more than this number of
    // assigned ltierals
    uint priBacktrackPickLimit;
    // number of variables that are pushed to dec_stack and
    // have the absolute priority over all others
    uint priUltimatePriVars; // */

    // clause parameters
    uint clRelSearchWindow;
    // misc
    uint faResizeTresh;
    uint cacheSize;
    uint maxCacheSize;
    uint cacheDelPerc;
    uint cacheDelLimitPtr;
    float cacheInflationRate;
    char priUpdatePolicy;
    char cacheUpdatePolicy;
    char impLongChains;
} params_t;

typedef struct var_t {
    uint fheap_idx;
    // priority queue node
    fheap_node_t *fnode;
    // decision level
    uint dlev;
    // number of scheduled instances of the variable
    uint sched_cnt;
    // priority counters
    uint pri[2];
    // antecedent
    cla_t *reason;
    // list of scheduled implied variables
    flexarray_t *uip_implied;
    // BCT anchors
    hc_anchor_t *hc[2];
} var_t;

#define CLSIDX(a,b) ((a << 1) + a + b)

typedef struct satm_t {
    // clause stacks
    stack_t **cls;
    // flags
    uintsh *typs;
    uint status;
    uint varnum;    // num of variables
    uint clnum;     // num of original clauses
    uint num_marked;// num of marked vars for learnDEC/UIP 
    uint learn1UIP;
    clock_t timeout;
    // flexarray used for passing conflict vars around
    flexarray_t *conflict;
    // temporary flexarray used for supercubing
    flexarray_t *tmp_fa;
    // progress array
    flexarray_t *progress;
    // temporary storage for constructed conflict clause
    cla_t *conflicting_cl;
    // used for storing implied variables which
    // need to be resolved in learnDEC
    stack_t *tmp_stack;
    // stores marked implied vars which need to
    // be unmarked when learnDEC is done
    stack_t *tounmark;
    stack_t *unit_cls;
    uint *learned_cls;
    // pointer into learned clauses cache
    uint lcls_ptr;
    int (*hfunc)(uint a, uint b);
    int (*init_hfunc)(uint a, uint b, uint c);
    /* TODO: delete...
    uint intermediate_added;
    uint intermediate_constructed;
    cla_t *intermediate;
    // */
#ifndef __GNUC__
    uint *tstack;
#endif
    hc_support_t *hcs;
    params_t *params;
    stats_t *stats;
    // implication queue
    bcpq_t *impq;
    // assignment queue
    assign_t *aq;
    uint fheap_ptr;
    uint fheap_sz;
    fheap_t **fheap;
    var_t *vt[];
} satm_t;

// ******************* DECLARATIONS  *****************************

inline uint hasCube(const satm_t *sm, const uint v);
inline void clearCube(satm_t *sm, const uint v);
inline void markCube(satm_t *sm, const uint v);
inline uintsh getTyp(const satm_t *sm, const uint v);
inline void setTyp(satm_t *sm, const uint v, const uintsh typ);
inline void setSign(satm_t *sm, const uint v, const uint sgn);
inline uint getSign(const satm_t *sm, const uint v);
inline uint flipPhase(const uint lit);
inline uintsh isAssigned(const satm_t *sm, const uint v);
inline void markAssigned(satm_t *sm, const uint v);
inline void clearAssigned(satm_t *sm, const uint v);
inline uint isBranchable(const satm_t *sm, const uint v);
inline uintsh isMarked(const satm_t *sm, const uint v);
inline void clearMark(satm_t *sm, const uint v);
inline void markVar(satm_t *sm, const uint v);
inline uintsh isAdded(const satm_t *sm, const uint v);
inline void clearAdded(const satm_t *sm, const uint v);
inline void addVar(const satm_t *sm, const uint v);
inline void setDecLev(satm_t *sm, const uint v, const uint dlev);
inline uint getDecLev(const satm_t *sm, const uint v);
inline void regReason(satm_t *sm, const uint v, cla_t *cl);
inline cla_t *getReason(const satm_t *sm, const uint v);
inline void unregReason(satm_t *sm, const uint v);
inline uintsh isBothDir(const satm_t *sm, const uint v);
inline void markBothDir(satm_t *sm, const uint v);
//inline void clearBothDir(satm_t *sm, const uint v);
inline char checkClSanity(satm_t *sm, cla_t *cl);
void periodicFunctions(satm_t *sm, char typ, int param);
cla_t *constructClFromUIP(satm_t *sm, uint uip);
void mark_vars(satm_t *sm, cla_t *cl, uint skip, uint dl);
uint learnUIP(satm_t *sm, cla_t *r, uint dlev);
void learnDEC(satm_t *sm, cla_t *r);
uint backtrackFlip(satm_t *sm);
uint resolve(satm_t *sm);
void hypercube(satm_t *sm);
uint decide(satm_t *sm);
void reset(satm_t *sm, const uint v);
inline uint isLitSat(satm_t *sm, uint lit);
uint compute(satm_t *sm, const uint lit);
satm_t *newManager(uint varnum, uint clnum, flexarray_t *elimvars,
        flexarray_t *depvars, flexarray_t *clusters, stats_t *stats);
void initManager(satm_t *sm, flexarray_t *cnf, flexarray_t *coe,
        uint **pri_mtx, clock_t tout);
void freeDeadClause(satm_t *sm, cla_t *cl);
void freeManager(satm_t *sm);
void setParams(satm_t *sm, params_t *params);
void search(satm_t *sm);
uint bcpqPut(satm_t *sm, uint lit, cla_t *cl);
inline uint bcpqGet(satm_t *sm);
void setManagerStatus(satm_t *sm, char status); 
int computeVarPriority(satm_t *sm, uint v);
inline uintsh isFixed(const satm_t *sm, const uint v);
inline void markFixed(satm_t *sm, const uint v, const uint sgn);
inline uint getSchedCnt(const satm_t *sm, const uint v);
inline void incSchedCnt(satm_t *sm, const uint inc, const uint v);
inline void decSchedCnt(satm_t *sm, const uint v);
uint isClDeletable(satm_t *sm, cla_t *cl);
void clausePriorityUpdate(satm_t *sm, cla_t *cla, char ctyp);
void add_CNF_cl(satm_t *sm, cla_t *c);
void add_CNF_lst(satm_t *sm, flexarray_t *hd);
void add_COE_cl(satm_t *sm, cla_t *c);
void add_COE_lst(satm_t *sm, flexarray_t *hd);
void init_lit_priorities(satm_t *sm, uint **mtx);
int declev_cmp(const uint data, const void *v1, const void *v2);
void addToCache(satm_t *sm, cla_t *c);
uint decideSign(satm_t *sm, uint pos, uint neg);
void resetSchedImplied(satm_t *sm, uint v);

// END OF DECLARATIONS

// WARRNING: this function depends on literal encoding
uint bcpqPut(satm_t *sm, uint lit, cla_t *cl) {
    bcpq_t *m = sm->impq;
    uint v = getVar(lit);
    // POS (10) if 1, NEG (01) if 0
    uint res, sgn = getPhase(lit) + 1; 
    assert(lit > 1 && lit < (sm->varnum + 1) << 1);
    assert(cl != NULL);
    assert(checkClSanity(sm, cl));
    res = writeTable(m->t, v, sgn);
    if (res == WRITE_SUCCESS) {
        m->a[m->widx++] = lit;
        m->reasons[v] = cl;
        return BCP_OK;
    } else if (res == sgn) {
        cla_t *reg_cl = m->reasons[v];
        if (getClSize(cl) < getClSize(reg_cl)) {
            m->reasons[v] = cl;
        }
        return BCP_EXISTS;
    } 
    while (m->ridx != m->widx) {
        uint c_lit = m->a[m->ridx++];
        if ((c_lit ^ lit) == 1) {
            regReason(sm, v, m->reasons[v]);
            break;
        }
    }
    // reset the bcp fifo
    bcpqReset(m);
    setDecLev(sm, v, declev(sm->aq));
    appendAQ(sm->aq, lit, declev(sm->aq), FALSE);
    return BCP_CONFLICT;
}  

inline uint bcpqGet(satm_t *sm) {
    uint lit, v;
    bcpq_t *m = sm->impq;
    cla_t *reason;
    assert(!bcpqEmpty(m));
    lit = m->a[m->ridx++];
    v = getVar(lit);
    reason = m->reasons[v];
    assert(v > 0 && v <= sm->varnum);
    assert(reason != NULL);
    assert(checkClSanity(sm, reason));
    regReason(sm, v, reason);
    return lit;
}

inline int computeVarPriority(satm_t *sm, uint v) {
    //assert(sm->vt[v] != NULL);
    return sm->hfunc(sm->vt[v]->pri[0],sm->vt[v]->pri[1]);
}

int getVarPriority(uint sm_ref, uint v) {
    return computeVarPriority((satm_t *)sm_ref, v);
}

inline uint getSchedCnt(const satm_t *sm, const uint v) {
    //assert(sm->vt[v] != NULL);
    return sm->vt[v]->sched_cnt;
}

inline void incSchedCnt(satm_t *sm, const uint inc, const uint v) {
    //assert(sm->vt[v] != NULL);
    sm->vt[v]->sched_cnt += inc;
}

inline void decSchedCnt(satm_t *sm, const uint v) {
    var_t *ve = sm->vt[v];
    assert(ve != NULL);
    if (ve->sched_cnt > 0) {
        ve->sched_cnt--;
    }
}

int revdeclev_cmp(const uint data, const void *v1, const void *v2) {
    satm_t *sm = (satm_t *)data;
    uint var1 = *(uint *)v1;
    uint var2 = *(uint *)v2;
	uint d1, d2;
    assert(var1 > 0 && var1 <= sm->varnum);
    assert(var2 > 0 && var2 <= sm->varnum);
    d1 = getDecLev(sm, var1);
    d2 = getDecLev(sm, var2);
    // smaller dec levels first
    return (int)d2 - d1;
}// */

int declev_cmp(const uint data, const void *v1, const void *v2) {
    satm_t *sm = (satm_t *)data;
    uint var1 = *(uint *)v1;
    uint var2 = *(uint *)v2;
	uint d1, d2;
    assert(var1 > 0 && var1 <= sm->varnum);
    assert(var2 > 0 && var2 <= sm->varnum);
    d1 = getDecLev(sm, var1);
    d2 = getDecLev(sm, var2);
    // smaller dec levels first
    return (int)d1 - d2;
}

/*
int sz_cmp(const void *ca, const void *cb) {
    cla_t *a = (cla_t *)(*((uint *)ca));
    cla_t *b = (cla_t *)(*((uint *)cb));
    return (int)getClSize(a) - getClSize(b);
}// */

inline uint hasCube(const satm_t *sm, const uint v) {
    assert(v > 0 && v <= sm->varnum);
    assert(sm->vt[v] != NULL);
    return sm->vt[v]->hc[0]->has_hcube;
}

inline void clearCube(satm_t *sm, const uint v) {
	var_t *ve;
	hc_anchor_t *ha;
    assert(v);
    ve = sm->vt[v];
    //assert(ve != NULL);
    ha = ve->hc[0];
    resetHcube(sm->hcs, ha);
    ha = ve->hc[1];
    if (ha->nd) {
		hc_node_t *nd = ha->nd;
        assert((uint)nd > 1);
        freeHcNode(sm->hcs, nd->nxt);
        freeHcNode(sm->hcs, nd->neg);
        ha->nd = nd->nxt = nd->neg = NULL;
    }
    resetHcSupport(sm->hcs);
}
    
inline void markCube(satm_t *sm, const uint v) {
	var_t *vt;
    assert(v > 0 && v <= sm->varnum);
    vt = sm->vt[v];
    //assert(vt != NULL);
    assert(vt->hc[0] != NULL);
    vt->hc[0]->has_hcube = TRUE;
}

inline void incNumLearned(satm_t *sm) {
    sm->stats->clNumLearned++;
}

inline void decNumLearned(satm_t *sm) {
    sm->stats->clNumLearned--;
}

inline uintsh getTyp(const satm_t *sm, const uint v) {
    assert(v > 0 && v <= sm->varnum);
    //assert(sm->vt[v] != NULL);
    return sm->typs[v];
}

inline void setTyp(satm_t *sm, const uint v, const uintsh typ) {
    assert(v > 0 && v <= sm->varnum);
    //assert(sm->vt[v] != NULL);
    sm->typs[v] = typ;
}

inline void setSign(satm_t *sm, const uint v, const uint sgn) {
#ifdef UGLY_HACKS
#ifdef DEBUG
    uintsh t, ta;
#endif
#endif
    assert(v > 0 && v <= sm->varnum);
    //assert(sm->vt[v] != NULL);
    assert(sgn < 2);
#ifdef UGLY_HACKS
#ifdef DEBUG
    t = getTyp(sm, v);
    sm->typs[v] ^= (-sgn ^ getTyp(sm, v)) & VALUE;
    ta = getTyp(sm, v);
    assert(!sgn || ta == (t | VALUE));
    assert(sgn || ta == (t & ~VALUE));
#else 
    sm->typs[v] ^= (-sgn ^ getTyp(sm, v)) & VALUE;
#endif
#else
    setTyp(sm, v, sgn ? getTyp(sm,v) | VALUE : getTyp(sm,v) & ~VALUE);
#endif
}

inline uint getSign(const satm_t *sm, const uint v) {
    assert(v > 0 && v <= sm->varnum);
    //assert(sm->vt[v] != NULL);
    return (uint)getTyp(sm, v) & VALUE;
}

inline uint flipPhase(const uint lit) {
    assert(lit);
    return lit ^ 1;
}

#ifdef DEBUG
inline void markLiteralNd(satm_t *sm, const uint v) {
    assert(v > 0 && v <= sm->varnum);
    //assert(sm->vt[v] != NULL);
    sm->typs[v] |= LITERAL_BCT_ND;
}

inline void markBranchNd(satm_t *sm, const uint v) {
    assert(v > 0 && v <= sm->varnum);
    //assert(sm->vt[v] != NULL);
    sm->typs[v] |= BRANCH_BCT_ND;
}

inline uintsh isLiteralNd(const satm_t *sm, const uint v) {
    assert(v > 0 && v <= sm->varnum);
    //assert(sm->vt[v] != NULL);
    return getTyp(sm, v) & LITERAL_BCT_ND;
}

inline uintsh isBranchNd(const satm_t *sm, const uint v) {
    assert(v > 0 && v <= sm->varnum);
    //assert(sm->vt[v] != NULL);
    return getTyp(sm, v) & BRANCH_BCT_ND;
}

inline void clearBranchNd(satm_t *sm, const uint v) {
    assert(v > 0 && v <= sm->varnum);
    //assert(sm->vt[v] != NULL);
    sm->typs[v] &= ~BRANCH_BCT_ND;
}
#endif

inline uintsh isFixed(const satm_t *sm, const uint v) {
    return getTyp(sm, v) & FIXED;
}

inline void markFixed(satm_t *sm, const uint v, const uint sgn) {
    assert(v > 0 && v <= sm->varnum);
    //assert(sm->vt[v] != NULL);
    assert(!isFixed(sm,v));
    sm->typs[v] |= FIXED;
    if (sgn) {
        sm->typs[v] |= FIXED_SIGN;
    }
}

inline uint getFixedSgn(satm_t *sm, const uint v) {
    assert(v > 0 && v <= sm->varnum);
    //assert(sm->vt[v] != NULL);
    assert(isFixed(sm,v));
    return TOBOOL((sm->typs[v] & FIXED_SIGN));
}

inline uintsh isAssigned(const satm_t *sm, const uint v) {
    assert(v > 0 && v <= sm->varnum);
    return getTyp(sm, v) & ASSIGNED;
}

inline void markAssigned(satm_t *sm, const uint v) {
    assert(v > 0 && v <= sm->varnum);
    //assert(sm->vt[v] != NULL);
    sm->typs[v] |= ASSIGNED;
}

inline void clearAssigned(satm_t *sm, const uint v) {
    assert(v > 0 && v <= sm->varnum);
    //assert(sm->vt[v] != NULL);
    sm->typs[v] &= ~ASSIGNED;
}

inline uint isBranchable(const satm_t *sm, const uint v) {
    assert(v > 0 && v <= sm->varnum);
    return !(getReason(sm, v) || isBothDir(sm, v));
}

inline void markDependant(satm_t *sm, const uint v) {
    assert(v > 0 && v <= sm->varnum);
    //assert(sm->vt[v] != NULL);
    sm->typs[v] |= DEPENDANT;
}

inline uintsh isDependant(const satm_t *sm, const uint v) {
    assert(v > 0 && v <= sm->varnum);
    return getTyp(sm, v) & DEPENDANT;
}

inline void markEliminated(satm_t *sm, const uint v) {
    assert(v > 0 && v <= sm->varnum);
    //assert(sm->vt[v] != NULL);
    sm->typs[v] |= ELIMINATED;
}
    
inline uintsh isEliminated(const satm_t *sm, const uint v) {
    assert(v > 0 && v <= sm->varnum);
    return getTyp(sm, v) & ELIMINATED;
}

inline uintsh isMarked(const satm_t *sm, const uint v) {
    assert(v > 0 && v <= sm->varnum);
    return getTyp(sm, v) & MARK;
}

inline void clearMark(satm_t *sm, const uint v) {
    assert(v != 0 && v <= sm->varnum);
    assert(sm->num_marked > 0);
    assert(isMarked(sm, v));
    //assert(sm->vt[v] != NULL);
    sm->num_marked--;
    sm->typs[v] &= ~MARK;
}

inline void markVar(satm_t *sm, const uint v) {
    assert(v != 0 && v <= sm->varnum);
    assert(sm->num_marked <= sm->varnum);
    assert(!isMarked(sm, v));
    //assert(sm->vt[v] != NULL);
    sm->num_marked++;
    sm->typs[v] |= MARK;
}

inline uintsh isAdded(const satm_t *sm, const uint v) {
    assert(v > 0 && v <= sm->varnum);
    return getTyp(sm, v) & ADDED;
}

inline void clearAdded(const satm_t *sm, const uint v) {
    assert(v > 0 && v <= sm->varnum);
    assert(isAdded(sm, v));
    //assert(sm->vt[v] != NULL);
    sm->typs[v] &= ~ADDED;
}

inline void addVar(const satm_t *sm, const uint v) {
    assert(v > 0 && v <= sm->varnum);
    assert(!isAdded(sm, v));
    //assert(sm->vt[v] != NULL);
    sm->typs[v] |= ADDED;
}

inline void setDecLev(satm_t *sm, const uint v, const uint dlev) {
    assert(v > 0 && v <= sm->varnum);
    //assert(sm->vt[v] != NULL);
    sm->vt[v]->dlev = dlev;
}

inline uint getDecLev(const satm_t *sm, const uint v) {
    assert(v > 0 && v <= sm->varnum);
    //assert(sm->vt[v] != NULL);
    return sm->vt[v]->dlev;
}

inline void regReason(satm_t *sm, const uint v, cla_t *cl) {
    assert(v > 0 && v <= sm->varnum);
    assert(checkClSanity(sm, cl));
    //assert(sm->vt[v] != NULL);
    sm->vt[v]->reason = cl;
}

inline cla_t *getReason(const satm_t *sm, const uint v) {
    assert(v > 0 && v <= sm->varnum);
    //assert(sm->vt[v] != NULL);
    return sm->vt[v]->reason;
}

inline void unregReason(satm_t *sm, const uint v) {
#ifdef DEBUG
    cla_t *r;
#endif
    assert(v > 0 && v <= sm->varnum);
    //assert(sm->vt[v] != NULL);
#ifdef DEBUG
    r = sm->vt[v]->reason;
    if (r) {
        assert(checkClSanity(sm, r));
    }
#endif
    sm->vt[v]->reason = NULL;
}

inline uintsh isBothDir(const satm_t *sm, const uint v) {
    assert(v > 0 && v <= sm->varnum);
    //assert(sm->vt[v] != NULL);
    return getTyp(sm, v) & BOTH_DIR;
}

inline void markBothDir(satm_t *sm, const uint v) {
    assert(v > 0 && v <= sm->varnum);
    //assert(sm->vt[v] != NULL);
    sm->typs[v] |= BOTH_DIR;
}

/*inline void clearBothDir(satm_t *sm, const uint v) {
    assert(v > 0 && v <= sm->varnum);
    //assert(sm->vt[v] != NULL);
    sm->typs[v] &= ~BOTH_DIR;
}// */

void init_lit_priorities(satm_t *sm, uint **mtx) {
    uint i;
    uint pri = sm->params->priInitial;
    static uint freqs[5] = {0, 0, 0, 0, 0};
    for (i = 1; i <= sm->varnum; i++) {
        uint l, negp, posp, bin;
        int score;
        var_t *vt = sm->vt[i];
        // skip it if it's eliminated or dependant variable 
        if (isEliminated(sm, i) || isDependant(sm, i)) continue;
        l = i << 1;
        negp = mtx[l][pri];
        bin = mtx[l][PRI_BIN_CLS];
        l++;
        posp = mtx[l][pri];
        vt->pri[0] = negp;
        vt->pri[1] = posp;
        bin += mtx[l][PRI_BIN_CLS];
        score = sm->init_hfunc(posp, negp, bin);
        freqs[vt->fheap_idx]++;
        fh_node_insert(sm->fheap[vt->fheap_idx], vt->fnode, -score);
    }
}

void add_COE_cl(satm_t *sm, cla_t *c) {
    uint j, sz = getClSize(c);
    assert(sz > 3 || sz == 0);
    assert(isEquiv(c));
    assert(!isLearned(c));
    if (sz == 0) return;
    assert(isEquiv(c));
    assert(checkClSanity(sm, c));
    for (j = 1; j < 3; j++) {
        //assert(sm->vt[c[j]] != NULL);
        stackPush(sm->cls[CLSIDX(c[j],2)], (Item)((uint)c | j));
    }
}

void add_CNF_lst(satm_t *sm, flexarray_t *hd) {
    uint *a = getFlexArray(hd);
    uint i, sz = getFlexArraySize(hd);
    for (i = 0; i < sz; i++) {
        cla_t *c = (cla_t *)a[i];
        add_CNF_cl(sm, c);
    }
}

void add_COE_lst(satm_t *sm, flexarray_t *hd) {
    uint *a = getFlexArray(hd);
    uint i, sz = getFlexArraySize(hd);
    for (i = 0; i < sz; i++) {
        cla_t *c = (cla_t *)a[i];
        add_COE_cl(sm, c);
    }
}

//TODO: put the most frequent events first
void periodicFunctions(satm_t *sm, char typ, int param) {
    stats_t *s = sm->stats;
    params_t *p = sm->params;
    switch (typ) {
    case EVNT_CONFLICT :
        s->cnfNum++;
        if ((s->cnfNum % p->priScalePeriod) == 0) {
            uint i,j;
            for (i = 1; i <= sm->varnum; i++) {
                var_t *vt = sm->vt[i];
                if (!vt || isAssigned(sm,i)) continue;
                for (j = 0; j < 2; j++) {
                    //assert(vt != NULL);
                    vt->pri[j] >>= 1;
                }
            }
        }
        break;
#ifdef COLLECT_STATS
    case EVNT_DECISION:
        s->decNum++;    
        s->decMaxLev = _max(param, (int)s->decMaxLev);
        break;
    case EVNT_DEC_FIX_LEARN:
        s->decNumFixedLearn++;
        break;
    case EVNT_ASSIGNMENT :
        s->decMaxAssigned = 
            _umax(s->decMaxAssigned, numAssignedAQ(sm->aq));
        break;
    case EVNT_UIP_CONSTRUCT:
        s->cnfUIPs++;
        s->cnfUIPCumulativeSz += (long)param;
        break;
    case EVNT_DEC_CONSTRUCT:
        s->cnfDECs++;
        s->cnfDECCumulativeSz += (long)param;
        break;
    case EVNT_NEW_HCUBE_BRANCHES:
        s->hcTotBranches += (long)param;
        break;
    case EVNT_NEW_HCUBE_NODES:
        s->hcNumHcubes += (param > 0);
        s->hcTotNodes += (long)param;
        break;
    case EVNT_USED_HCUBE_BRANCH:
        s->hcUseBranches ++;
        break;
    case EVNT_USED_HCUBE_NODE:
        s->hcUseNodes ++;
        break;
    case EVNT_HCUBE_INTERSECTION:
        s->hcIntersections += (long)param;
        break;
    case EVNT_HC_INTERS_CONF:
        s->hcIntersectConflicts++;
        break;
    case EVNT_SCHEDULE_IMPS:
        s->impBlocksScheduled++;
        s->impTotalBlockSz += (long)param;
        break;
    case EVNT_IMP_CHAIN:
        s->impNumChains++;
        s->impLongestChain = _max(param, (int)s->impLongestChain);
        break;
    case EVNT_IMPLICATION:
        s->impNumImps++;
        break;
    case EVNT_RESTART:
        s->restarts++;
        break;
#endif
    default:
        break;
    }
}

void add_CNF_cl(satm_t *sm, cla_t *c) {
    assert(getClSize(c) > 0);
    assert(!isEquiv(c));
    assert(checkClSanity(sm, c));
    if (isUnitCl(c)) {
        assert(isLearned(c));
        incNumLearned(sm);
        stackPush(sm->unit_cls, (Item)c);
        periodicFunctions(sm, EVNT_DEC_FIX_LEARN, 0);
    } else {
        uint j;
        for (j = 1; j < 3; j++) {
            uint l = c[j];
			stack_t *s;
            uint v = getVar(l);
            //assert(sm->vt[v] != NULL);
            s = sm->cls[CLSIDX(v,getPhase(l))];
            stackPush(s, (Item)((uint)c | j));
        }
        if (isLearned(c)) {
            incNumLearned(sm);
        }
    }
}

inline void quickPriUpdate(satm_t *sm, const uint v, const uint idx, 
        const char typ, const uint inc) {
    var_t *vt;
    uint tot_inc;
    if (typ != sm->params->priUpdatePolicy || isBranchable(sm, v)) {
        return;
    }
    //assert(sm->vt[v] != NULL);
    vt = sm->vt[v];
    tot_inc = (sm->params->priIncrement + TOBOOL(!getReason(sm,v)) + inc);
    if (vt->pri[idx] + tot_inc > vt->pri[idx]) {
        vt->pri[idx] += tot_inc;
    }
}

void clausePriorityUpdate(satm_t *sm, cla_t *c, char ctyp) {
    char typ = sm->params->cacheUpdatePolicy;
    if (typ != ctyp || !isLearned(c) || isUnitCl(c)) {
        return;
    }
    incUsage(c);
}

inline void sortConflict(satm_t *sm, flexarray_t *conf) {
    uint *a = getFlexArray(conf);
    uint sz = getFlexArraySize(conf);
    cmp_tqsort(a, sz, 8, (uint)sm, revdeclev_cmp);
    if (sz >= 2) {
        swap((int *)a, 0, 1);
    }
}

cla_t *constructClFromUIP(satm_t *sm, uint uip) {
    cla_t *xc;
    uint sz;
    flexarray_t *c = sm->conflict;
    appendToFlexArray(c, uip);
    sortConflict(sm, c);
    sz = getFlexArraySize(c);
    xc = getNewClause(getFlexArray(c), sz);
    assert(getClSize(xc) == sz + 1);
    // uip must end up in second place
    assert(xc[_min(2, (int)sz)] == uip);
    return xc;
} 

void mark_vars(satm_t *sm, cla_t *c, uint skip, uint dl) {
    uint cnt = 0, i = getClSize(c) - 1, *sp;
#ifdef __GNUC__
    uint stack[i];
#else
    uint *stack = sm->tstack;
#endif
    sp = stack;

    assert(dl != 0);
    assert(checkClSanity(sm, c));
    assert(i > 0);
   
    if (isEquiv(c)) { 
        do {
            uint v = c[i];
            sp[0] = v;
            assert((isMarked(sm, v) || v == skip) < 2);
            sp += (isMarked(sm, v) || v == skip) ^ 1;
        } while (--i);
    } else {
        do {
            uint v = getVar(c[i]), sgn = LSB(c[i]);
            uint inc = (uint)(isBinCl(c)) * sm->params->priBinBoost;
            if (cnt++ < sm->params->priLitsUpdated) {
                quickPriUpdate(sm, v, sgn, PRI_UIP_CONSTRUCTION, inc);
            }
            quickPriUpdate(sm, v, sgn, PRI_UPDATE_ALL, inc);
            sp[0] = v;
            assert((isMarked(sm, v) || v == skip) < 2);
            sp += (isMarked(sm, v) || v == skip) ^ 1;
        } while (--i);
    }
    if (sp == stack) {
        return;
    }
    do {
		uint v, vdl;
        sp--;
        v = sp[0];
        // there can be only one unassigned variable in the 
        // conflict - the conflicting variable, all others
        // must be assigned
        assert(isAssigned(sm, v) || 
                (!isAssigned(sm, v) && getDecLev(sm, v) == dl));
        vdl = getDecLev(sm, v);
        if (vdl == dl) {
            markVar(sm, v);
        } else if (!(isAdded(sm, v) || isFixed(sm, v))) {
            uint *fa;
            assert(!isMarked(sm, v));
            assert(getDecLev(sm, v) < dl);
            assert(isAssigned(sm, v));
            assert(v > 0 && v <= sm->varnum);
            // WARRNING: literals must be formed before adding
            // the clause to the database! according to:
            // (var << 1) | (current_assignment ^ 1)
            appendToFlexArray(sm->conflict, v);
            addVar(sm, v);
            fa = getFlexArray(sm->conflict);
        }
    } while (sp != stack);
}

/*
#define INTERMEDIATE_CUT_PERC 50
#define INTERMEDIATE_DIF_PERC 20
#define INTERMEDIATE_ADD_PERC (INTERMEDIATE_CUT_PERC + INTERMEDIATE_DIF_PERC)
void addIntermediate(satm_t *sm, uint resolutions, uint *sp, uint dlev) {
    uint avg = sm->stats->resAvgNum;
    uint cut_res_num = (INTERMEDIATE_CUT_PERC * avg) / 100;
    if (sm->intermediate_constructed) {
        uint add_res_num, csz;
        cla_t *c = sm->intermediate;
        add_res_num = (INTERMEDIATE_ADD_PERC * avg) / 100;
        if (resolutions < add_res_num || (add_res_num - cut_res_num) < 3) {
            return;
        } 
        csz = getClSize(c);
        assert(csz > 2);
        csz--;
        cmp_tqsort(&(c[1]), csz, 8, (uint)sm, revdeclev_cmp);
        do {
            uint xv = c[csz];
            //assert(sm->vt[xv]);
            assert(isAssigned(sm,xv));
            c[csz] = makeLit(xv, getSign(sm, xv) ^ 1);
        } while (--csz);
        add_CNF_cl(sm, c);
        addToCache(sm, c);
        sm->intermediate_added = TRUE;
        sm->intermediate_constructed = FALSE;
        sm->intermediate = NULL;
    } else if (!sm->intermediate_added && resolutions > 2) {
        if (avg && resolutions > cut_res_num) {
            flexarray_t *conf = sm->conflict;
            uint sz_new, sz_orig = getFlexArraySize(conf);
            uint *stack = sm->aq->a;
            cla_t *c;
            do {
                sp--;
                uint v = getVar(sp[0]);
                if (isMarked(sm, v)) {
                    appendToFlexArray(conf, v);
                } else if (getDecLev(sm, v) < dlev) {
                    break;
                }
            } while (sp != stack);
            sz_new = getFlexArraySize(conf);
            assert(sz_new > sz_orig);
            c = getNewClause(getFlexArray(conf), sz_new);
            conf->idx = sz_orig;
            markLearned(c);
            sm->intermediate = c;
            sm->intermediate_constructed = TRUE;
        }
    }
}// */

uint learnUIP(satm_t *sm, cla_t *r, uint dlev) {
    uint *stack = sm->aq->a, *sp = &(sm->aq->a[sm->aq->aidx]);
    uint uip = 0, confsz;
    /*
    uint resolutions = 1;
    assert(!sm->intermediate_added);
    assert(!sm->intermediate_constructed);
    // */

    assert(r != NULL);
    assert(!isUnitCl(r));
    assert(checkClSanity(sm, r));
    assert(sm->num_marked == 0);
    assert(getFlexArraySize(sm->conflict) == 0);
    assert(sp != stack);

    mark_vars(sm, r, 0, dlev);
    assert(sm->num_marked > 0);
    do {
		uint v;
        sp--;
        v = getVar(sp[0]);
        if (isMarked(sm, v)) {
            clearMark(sm, v);
            if (sm->num_marked) {
                cla_t *x = getReason(sm, v);
                //resolutions++;
                assert(x != NULL);
                mark_vars(sm, x, v, dlev);
                clausePriorityUpdate(sm, x, UIP_CONFLICTING_CLS);
                //addIntermediate(sm, resolutions, sp, dlev);
            } else { // uip found
                assert(!isAdded(sm, v));
                // WARRNING: uip literal must be formed before
                // adding learned clause to the database
                uip = v;
                break;
            }
        } else if (getDecLev(sm, v) < dlev) {
            assert(sm->num_marked == 0);
            break;
        }
    } while (sp != stack);
    
    confsz = getFlexArraySize(sm->conflict);
    if (confsz) {
        uint *conf = getFlexArray(sm->conflict);
        do {
            confsz--;
            clearAdded(sm, conf[confsz]);
        } while (confsz);
    }
    
    assert(uip != 0);
    assert(sm->num_marked == 0);
    /*
    periodicFunctions(sm, EVNT_IMP_RESOLUTION, resolutions);
    sm->intermediate_added = FALSE;
    if (sm->intermediate) {
        assert(sm->intermediate_constructed);
        sm->intermediate_constructed = FALSE;
        free(sm->intermediate);
        sm->intermediate = NULL;
    }// */
    return uip;
}

void learnDEC(satm_t *sm, cla_t *r) {
    stack_t *s = sm->tmp_stack;

    assert(stackEmpty(s));
    assert(r != NULL);
    assert(sm->num_marked == 0);
    assert(getFlexArraySize(sm->conflict) <= 1);

    stackPush(s, (uint)r);
    
    do {
        cla_t *c = (cla_t *)stackPop(s);
        uint i = getClSize(c) - 1, *sp;
        // local stack
#ifdef __GNUC__
        uint stack[i];
#else
        uint *stack = sm->tstack;
#endif
        sp = stack;
        clausePriorityUpdate(sm, c, DEC_CONFLICTING_CLS);
        // push all not marked and not added vars to local stack
        if (isEquiv(c)) {
            do {
                uint v;
                v = c[i];
                assert(!(isAdded(sm, v) && isMarked(sm, v)));
                sp[0] = v;
                // if it's marked, must have a reason
                assert(!isMarked(sm, v) || getReason(sm, v));
                // add variable to stack only if it's not marked
                // or added
                sp += !(isMarked(sm, v) || isAdded(sm, v));
            } while (--i);
        } else {
            do {
                uint v;
                v = getVar(c[i]);
                assert(!(isAdded(sm, v) && isMarked(sm, v)));
                //quickPriUpdate(sm, v, LSB(c[i]), PRI_DEC_CONSTRUCTION, 0);
                sp[0] = v;
                // if it's marked, must have a reason
                assert(!isMarked(sm, v) || getReason(sm, v));
                // add variable to stack only if it's not marked
                // or added
                sp += !(isMarked(sm, v) || isAdded(sm, v));
            } while (--i);
        }
        if (sp == stack) continue;
        do {
            uint v;
            cla_t *nr;
            sp--;
            v = sp[0];
            assert(!(isMarked(sm, v) || isAdded(sm, v)));
            nr = getReason(sm, v);
            if (nr) {
                markVar(sm, v);
                stackPush(s, (uint)nr);
                stackPush(sm->tounmark, v);
            } else {
                assert(isAssigned(sm, v));
                appendToFlexArray(sm->conflict, v);
                addVar(sm, v);
            }
        } while (sp != stack);
    } while (!stackEmpty(s));

    while (!stackEmpty(sm->tounmark)) {
        clearMark(sm, (uint)stackPop(sm->tounmark));
    }
    assert(sm->num_marked == 0);
    {
        uint *conf = getFlexArray(sm->conflict);
        uint fsz, i;
        fsz = i = getFlexArraySize(sm->conflict);
        if (i == 0) return;
        do {
            clearAdded(sm, conf[--i]);
        } while (i); 
        periodicFunctions(sm, EVNT_DEC_CONSTRUCT, (int)fsz);
        cmp_qsort(conf, fsz, 4, ucmp);
    }
}

void scheduleHcube(satm_t *sm, hc_node_t *n) {
    hc_node_t *x = n;
    while (x && x != TERM_ND) {
#ifdef DEBUG
        uint var = getVar(x->var);
        assert(var > 0 && var <= sm->varnum);
#endif
        // if it's an ordinary node, schedule it for computation
        if (!hc_is_bnode(x)) {
            scheduleAQ(sm->aq, x->var, SCHD_HC_STEM);
            x = x->nxt;
        } else {
            assert((uint)x->nxt > 1 || (uint)x->neg > 1);
            // the sign of branch nodes is chosen when (and if)
            // the node is reached
            scheduleAQ(sm->aq, x->var, (uint)x);               
            break;
        }
    }
}

void scheduleImplications(satm_t *sm, flexarray_t *fa) {
    uint *a = getFlexArray(fa);
    uint sz = getFlexArraySize(fa);
    if (sz) {
        uint i = 0;
        assert(sz % 2 == 0);
        do {
#ifdef DEBUG            
            uint res = bcpqPut(sm, a[i], (cla_t *)a[i + 1]);
            assert(res != BCP_CONFLICT);        
#else
            (void)bcpqPut(sm, a[i], (cla_t *)a[i + 1]);
#endif
            i += 2;
        } while (i < sz);
        periodicFunctions(sm, EVNT_SCHEDULE_IMPS, (int)(sz >> 1));    
    }
}

void hcbCleanup(satm_t *sm, uint var, uint ph) {
    hc_anchor_t *ha = sm->vt[var]->hc[1];
    hc_node_t *nd, *bnd = ha->nd;
    assert(!bnd || hc_is_bnode(bnd));
    if (!bnd) {
        return;
    } else if (ph) {
        nd = bnd->nxt;
        // free explored branch
        freeHcNode(sm->hcs, nd);
        ha->nd = TRIM(bnd->neg);
    } else {
        nd = bnd->neg;
        freeHcNode(sm->hcs, nd);
        ha->nd = TRIM(bnd->nxt);
    }
    bnd->nxt = bnd->neg = NULL;
    resetHcSupport(sm->hcs);
}

// puts new decision and supercubed variables to todo FIFO,
// and returns new decision level. backtrackFlip will also
// set basic flags for the new decision.
//
// backtrackFlip (the same as decide) sets only BOTH_DIR,
// DECISION flags, and decision level, flags for s/h-cubed
// and implied variables must be set by higher level funcs
uint backtrackFlip(satm_t *sm) {
    uint imp_chain_len = 0, new_dlev;
    resetScheduledAQ(sm->aq);

    while (1) {
        uint lit = popAQ(sm->aq), v;
        if (lit == 0) {
            new_dlev = 0; // problem is UNSAT
            break;
        }
        v = getVar(lit);
        assert(getDecLev(sm, v) > 0);
        if (!isBothDir(sm, v) && hasCube(sm, v)) {
            var_t *vt = sm->vt[v];
            new_dlev = getDecLev(sm, v);
            // flipped var must be DECISION
            assert(!getReason(sm, v));
            // and the sign of literal on assignment queue
            // must match the sign of variable in var table
            assert(getPhase(lit) == getSign(sm, v));
            hcbCleanup(sm, v, getPhase(lit));
            periodicFunctions(sm, EVNT_HCUBE_INTERSECTION, 
                (uint)vt->hc[0]->nd > 1 && (uint)vt->hc[1]->nd > 1);
            if (path_intersect(sm->hcs, vt->hc)) { // conflict
                markBothDir(sm, v);
                reset(sm, v);
                periodicFunctions(sm, EVNT_HC_INTERS_CONF, 0);
                periodicFunctions(sm, EVNT_IMP_CHAIN, (int)imp_chain_len);
                imp_chain_len = 0;
                continue;
            }
            reset(sm, v);
            // set minimum flags and decision level
            // as flipped variable will be evaluated
            // for sure... (this is not guaranteed for
            // supercubed variables, as we might reach
            // a conflict before reaching any of 
            // supercubed variables)
            markBothDir(sm, v);
            setDecLev(sm, v, new_dlev);
            // put it to todo FIFO 
            scheduleAQ(sm->aq, flipPhase(lit), SCHD_DECISION);
            // and put its supercubed vars to todo FIFO
            // variables with DECISION flag on the todo
            // FIFO are treated as decisions, all others
            // as super/hyper-cubed variables
            //assert(vt != NULL);
            periodicFunctions(sm, EVNT_NEW_HCUBE_BRANCHES, 
                (int)vt->hc[0]->num_bnodes);
            periodicFunctions(sm, EVNT_NEW_HCUBE_NODES,
                (int)vt->hc[0]->num_nodes);        
            scheduleHcube(sm, vt->hc[0]->nd);
            scheduleImplications(sm, vt->uip_implied);
            assert(numScheduledAQ(sm->aq) >= 1);
            break; 
        } else {
            cla_t *r = getReason(sm, v);
            imp_chain_len += (r != NULL);
            reset(sm, v);
            if (!r) {
                periodicFunctions(sm, EVNT_IMP_CHAIN, (int)imp_chain_len);
                imp_chain_len = 0;
            }
        }
    }
    periodicFunctions(sm, EVNT_IMP_CHAIN, (int)imp_chain_len);
    return new_dlev; 
}

void scheduleImplied(satm_t *sm, uint uiplit, uint uipcl, uint uipbdl) {
    uint i, dl = declev(sm->aq), schedcnt = 0, uipv = getVar(uiplit);
    assert(!isUnitCl((cla_t *)uipcl) || uipbdl == 0);
    if (!getReason(sm, uipv)) {
        // if uip is a decision variable, don't schedule it
        // as implied variable for the last dec level, as we
        // would run into the same implied var after flipping 
        // uip dec var
        dl--;
    }
    for (i = uipbdl + 1; i <= dl; i++) {
        uint v = getDecVar(sm->aq, i);
        flexarray_t *fa;
        assert(isAssigned(sm, v));
        assert(!getReason(sm, v));
        assert(getDecLev(sm, v) == i);
        if (!isBranchable(sm, v)) {
            continue;
        }
        // can't be a variable eliminated in preprocessing
        //assert(sm->vt[v] != NULL);
        // can't be implied variable
        assert(!getReason(sm, v));
        fa = sm->vt[v]->uip_implied;
        // at this point we also detect conflicts, and
        // set all branchable vars on higher levels than the conflict
        // level to unbranchable state!
        if (bfsearchFlexArray(fa, uiplit ^ 1, 2) >= 0) {
            uint j;
            for (j = i; j <= declev(sm->aq); j++) {
               uint x = getDecVar(sm->aq, j);
               if (!isBranchable(sm, x)) continue;
               assert(!isBothDir(sm, x));    
               markBothDir(sm, x);
            }
        } else {  
            appendToFlexArray(fa, uiplit);
            assert(uipcl);
            appendToFlexArray(fa, uipcl);
            schedcnt++;
        }
    }
    incSchedCnt(sm, schedcnt, uipv);
}

uint isClDeletable(satm_t *sm, cla_t *cl) {
    uint i, l[2], v[2];
    uint ass[2], sat[2];
    assert(!isUnitCl(cl));
    for (i = 0; i < 2; i++) {
        l[i] = cl[i+1];
        v[i] = getVar(l[i]);
        ass[i] = (uint)isAssigned(sm, v[i]);
        sat[i] = ass[i] && isLitSat(sm, l[i]);
    }

    // let one of the watched literals be assigned, but 
    // not satisfying the clause
    // than the clause is potentially registered
    // as a reason if:
    // 1) the second watched lit is not assigned, meaning
    // that it is a conflicting lit which never got
    // assigned. it must be a conflicting lit according to
    // watched lits invariant.
    // 2) if the var is assigned and the clause is registered
    // as the reason for its implication
    // 3) the variable corresponding to the second watched
    // lit has been scheduled as implication somewhere, meaning
    // that any clause might be registered as a reason
    if (ass[0] && !sat[0] && (!ass[1] || getReason(sm,v[1])==cl ||
                getSchedCnt(sm,v[1]) > 0)) return FALSE;
    if (ass[1] && !sat[1] && (!ass[0] || getReason(sm,v[0])==cl ||
                getSchedCnt(sm,v[0]) > 0)) return FALSE;

    return TRUE;
}

void addToCache(satm_t *sm, cla_t *c) {
    uint csz = sm->params->cacheSize;
    uint *ca = sm->learned_cls;
    if (sm->lcls_ptr < csz) {
        ca[sm->lcls_ptr++] = (uint)c;
    } else { // cache full
        uint limit;
        uint i = 0;
        uint cmp_thold;
        assert(csz == sm->lcls_ptr);
        limit = (sm->params->cacheDelPerc * csz) / 100;
        
        // TODO: set this as a parameter
        cmp_thold = 12;
        cmp_tqsort(ca, csz, 4, cmp_thold, cl_lucmp);

        do {
            cla_t *x = (cla_t *)ca[i]; 
            assert(x != NULL);
            if (isClDeletable(sm, x)) {
                freeDeadClause(sm, x);
                ca[i] = 0;
                limit--;
            }
            i++;
        } while (limit && i < csz);
        // compaction pass
        {
            int last_slot = -1;
            for (i = 0; i < csz; i++) {
                uint x = ca[i];
                if (x) {
                    cla_t *xc = (cla_t *)x;
                    scaleUsage(xc);
                    if (last_slot >= 0) {
                        assert(i != (uint)last_slot);
                        ca[last_slot++] = x;
                    }
                } else if (last_slot < 0) {
                    last_slot = (int)i;
                }
            }
            assert(last_slot >= 0);
            sm->lcls_ptr = _umax((uint)last_slot, (uint)0);
        }
        { 
            float rate = sm->params->cacheInflationRate;
            if (rate > 1.0) {
                uint new_size = (uint)(rate * csz);
                uint max_size = sm->params->maxCacheSize;
                assert(new_size > csz);
                sm->learned_cls = (uint *)xrealloc((char *)ca, 
                        new_size * sizeof(uint));
                sm->params->cacheSize = new_size;
                rate = (new_size > max_size) ? 1.0f : maxf(1.1f, rate - 0.08f);
                sm->params->cacheInflationRate = rate;
            }
        }
        sm->learned_cls[sm->lcls_ptr++] = (uint)c;
    }
}

#ifdef DEBUG
uint getMaxDlev(satm_t *sm, cla_t *c) {
    uint v[2] = {c[1], c[2]};
    if (!isEquiv(c)) {
        v[0] = getVar(v[0]);
        v[1] = getVar(v[1]);
    }
    return _umax(getDecLev(sm, v[0]), getDecLev(sm, v[1]));
}    
#endif

uint resolve(satm_t *sm) {
    uint *r = sm->conflicting_cl;
    assert(r != NULL);
    assert(checkClSanity(sm, r));
    assert(!isUnitCl(r));
    periodicFunctions(sm, EVNT_ASSIGNMENT, 0);
    if (sm->learn1UIP) {
        cla_t *lc;
        uint uip, lc_sz, sgn;
        assert(getMaxDlev(sm, r) == declev(sm->aq));
        uip = learnUIP(sm, r, declev(sm->aq));
        assert(uip > 0 && uip <= sm->varnum);
        assert(isAssigned(sm, uip));
        lc = constructClFromUIP(sm, uip);
        markEquiv(lc);
        resetFlexArray(sm->conflict);
        learnDEC(sm, lc);
        clearEquiv(lc);
        markLearned(lc);
        lc_sz = getClSize(lc);
        sgn = getSign(sm, uip) ^ 1;
        {   // schedule uip
            uint uipbdl = (isUnitCl(lc)) ? 0 : getDecLev(sm, lc[1]);
            uint uiplit = makeLit(uip, sgn);
            scheduleImplied(sm, uiplit, (uint)lc, uipbdl);
        }
        assert(lc_sz > 1);
        {
            uint idx = lc_sz - 1;
            do {
                uint xv = lc[idx];
                lc[idx] = makeLit(xv, getSign(sm, xv) ^ 1);
            } while (--idx);
        }
        add_CNF_cl(sm, lc);
        periodicFunctions(sm, EVNT_UIP_CONSTRUCT, (int)lc_sz);
        if (lc_sz > 2) {
            addToCache(sm, lc);
        } else {
            markFixed(sm, uip, sgn);
        }
    } else {
        learnDEC(sm, r);
    }
    assert(stackEmpty(sm->tmp_stack));
    sm->conflicting_cl = NULL;
    hypercube(sm);
    resetFlexArray(sm->conflict);
    periodicFunctions(sm, EVNT_CONFLICT, 0);
    sm->learn1UIP = TRUE;

    assert(bcpqEmpty(sm->impq));

    if (backtrackFlip(sm) == 0) {
        // problem is UNSAT
        return CONFLICT;
    } else {
        assert(numScheduledAQ(sm->aq) > 0);
        return NO_CONFLICT;
    }
}

/*
void copyNonLiteralDecisions(satm_t *sm, uint *a, uint sz, 
        flexarray_t *f) {
    uint i = 0;
    assert(getFlexArraySize(f) == 0);
    while (i < sz) {
        uint lit = a[i++];
        uint v = getVar(lit);
        if (!isLiteralNd(sm, v)) {
            appendToFlexArray(f, lit);
        }
    }
}// */

void hypercube(satm_t *sm) {
    flexarray_t *cert = sm->conflict;
    flexarray_t *tmp = partialInitFlex();
    uint *a = getFlexArray(cert);
    uint sz = getFlexArraySize(cert);
    if (sz) {
        uint i;
#ifdef DEBUG
        uint x;
#endif
        cmp_tqsort(a, sz, 8, (uint)sm, declev_cmp);
#ifdef DEBUG
        // check that the array if properly sorted by declev
        for (i = 0; i < sz; i++) {
            if (i > 0) {
                assert(getDecLev(sm, x) <= getDecLev(sm, a[i]));        
            } 
            x = a[i];
        }
#endif
        sz--;
        i = sz;
        do {
            uint v = a[i];
            uint elnum = sz - i;
            // the second condition is necessary to skip the implication
            // node (with flipped phase) that is added to the
            // certificate in the case of conflict between an
            // implication and a BCT literal node
            if (!isBothDir(sm, v) && !getReason(sm, v)) {
                var_t *vt = sm->vt[v];
                insertInFlex(tmp, &(a[i + 1]), elnum);
                common_path(sm->hcs, vt->hc[0], tmp);
            }
            if (getReason(sm, v)) {
                a[i] = makeLit(v, getSign(sm, v) ^ 1);
            } else {
                a[i] = makeLit(v, getSign(sm, v));
            }
        } while (i--);
    }
    partialFreeFlex(tmp);         
}

uint decideSign(satm_t *sm, uint pos, uint neg) {
    return (sm->params->impLongChains) ? (pos < neg) :
        (neg < pos);
}
    
#ifdef RESTARTS
// measure the progress of the solver
uint getReached(satm_t *sm, uint limit) {
    uint i = 1, dl = _min(limit, declev(sm->aq));
    while (i <= dl) {
        if (isBothDir(sm, getDecVar(sm->aq, i))) {
            return i;
        }
        i++;
    }
    return UINT_MAX;
}

uint no_progress(satm_t *sm) {
    flexarray_t *f = sm->progress;
    uint last, *a, sz = getFlexArraySize(f);
    if (sz < 4) { 
        return FALSE;
    } 
    a = getFlexArray(f);
    last = a[sz - 2];
    if (last > a[sz - 4]) {
        return TRUE;
    } else if (last == a[sz - 4]) {
        if (pseudorand(61) == 1) {
            return TRUE;
        } else if (last < 5 && pseudorand(51) == 1) {
            return TRUE;
        } else if (last >= 5 && last < 10 && pseudorand(31) == 1) {
            return TRUE;
        } else if (last >= 10 && last < 50 && pseudorand(21) == 1) {
            return TRUE;
        } else if (last >= 50 && pseudorand(11) == 1) {
            return TRUE;
        } else if (a[sz - 1] > a[sz - 3]) {
            return FALSE;
        } else {
            return TRUE; 
        }
    }
    return FALSE;
}

uint progress_status(satm_t *sm, uint reached, uint limit) {
    uint bd = 0, i = reached, dl = _min(reached + limit, declev(sm->aq));
    while (i <= dl) {
        if (isBothDir(sm, getDecVar(sm->aq, i))) {
            bd += (1 << (limit - (i - reached)));
        }
        i++;
    }
    return bd;
}
#endif
           
// decide will just make a decision, but won't mark any flags
uint decide(satm_t *sm) {
    uint l = 0, v = 0;
    // WARRNING: restarts are just an experimental patch for SAT
    // 2005 competition and haven't been thoroughly tested
#ifdef RESTARTS
    long conflicts = sm->stats->cnfNum;
    static long period = 2000;
    static long scheduled_restart = 2000;
    static uint limit = 26;
    static uint adaptations = 0;
#endif

    // if we're making a new decision, that means
    // that both implication queue and todo queue
    // (containing decision var and its supercubed
    // vars) must be empty
    assert(bcpqEmpty(sm->impq));
    // also, todo queue (containing decisions and
    // s/h-cubed vars) must be empty
    assert(numScheduledAQ(sm->aq) == 0);

#ifdef RESTARTS
    // restarts
    if (conflicts >= scheduled_restart) {
        uint reached = getReached(sm, UINT_MAX);
        appendToFlexArray(sm->progress, reached);
        appendToFlexArray(sm->progress, progress_status(sm, reached, limit));

        if (no_progress(sm)) {
            uint i, lit = popAQ(sm->aq);
            adaptations++;

            if (adaptations >= 2 && (adaptations & 1) == 0) {
                resetFlexArray(sm->progress);
            }

            if (adaptations == 2) {
                sm->params->impLongChains = FALSE;
            } else if (adaptations == 4) {
                sm->params->impLongChains = TRUE;
                sm->hfunc = hfunc3;
                sm->params->priLitsUpdated = 6;
            } else if (adaptations == 6) {
                sm->hfunc = hfunc6;
                sm->params->priBinBoost = 1; 
                sm->params->priUpdatePolicy = PRI_UPDATE_ALL;
                sm->params->priLitsUpdated = 4;
            } else if (adaptations == 8) {
                limit = _max(8, limit - 2);
                sm->hfunc = hfunc1;
                sm->params->priBinBoost = 2; 
                sm->params->priUpdatePolicy = PRI_UIP_CONSTRUCTION;
                adaptations = 0;
                period <<= 1;
            }
            scheduled_restart = conflicts + period;
            for (i = 0; i <= sm->fheap_sz; i++) {
                while(fh_delete_min(sm->fheap[i]));
            }
            while (lit) {
                uint rv = getVar(lit);
                clearCube(sm, rv);
                sm->typs[rv] &= (FIXED | DEPENDANT | FIXED_SIGN | ELIMINATED);
                setDecLev(sm, rv, 0);
                unregReason(sm, rv);
                resetSchedImplied(sm, rv);
                lit = popAQ(sm->aq);
            }
            resetScheduledAQ(sm->aq);
            resetAQ(sm->aq);
            bcpqReset(sm->impq);
            periodicFunctions(sm, EVNT_RESTART, 0);
            {
                uint scale, max_pri = 0;
                for (i = 1; i <= sm->varnum; i++) {
                    var_t *vt = sm->vt[i];
                    uint j;
                    if (!vt || isDependant(sm, i)) {
                        continue;
                    } else if (isFixed(sm, i)) {
                        uint xlit = (i << 1) | getFixedSgn(sm, i);
                        scheduleAQ(sm->aq, xlit, SCHD_HC_STEM);
                    }
                    for (j = 0; j < 2; j++) {
                        max_pri = _umax(max_pri, vt->pri[j]);
                    }
                }
                scale = (uint)_max(0, (int)approx_log2(max_pri) - 12);
                for (i = 1; i <= sm->varnum; i++) {
                    var_t *vt = sm->vt[i];
                    uint j;
                    if (!vt || isDependant(sm, i)) {
                        continue;
                    }
                    for (j = 0; j < 2; j++) {
                        vt->pri[j] >>= scale;
                    }
                    fh_node_insert(sm->fheap[vt->fheap_idx], vt->fnode, 
                        -computeVarPriority(sm, i));
                }
            }
            if (numScheduledAQ(sm->aq) > 0) {
                uint dummy;
                (void)getScheduledAQ(sm->aq, &l, &dummy);
                v = getVar(l);
                goto decision_made;
            }
        } else {
            scheduled_restart = conflicts + period;
        }
    }
#endif
    
    // get max priority var
    do {
        uint fptr = sm->fheap_ptr;
        uint pos, neg;

        do {
            v = (uint)fh_delete_min(sm->fheap[fptr]);
            if (v == 0) {
                fptr++;
                if (fptr < sm->fheap_sz) {
                    sm->fheap_ptr = fptr;
                } else {
                    return 0;
                }
            }
        } while (v == 0);

        //assert(sm->vt[v] != NULL);
        pos = sm->vt[v]->pri[0];
        neg = sm->vt[v]->pri[1];
        l = v << 1;
        l += decideSign(sm,pos,neg);
    } while (v && isAssigned(sm, v));

    // if the priority queue is empty, return 0
    // - all the variables are satisfied
    //if (!v) return 0;

#ifdef RESTARTS
decision_made:    
#endif
    // returned variable must be unassigned
    assert(isAssigned(sm, v) == 0);
    setDecLev(sm, v, declev(sm->aq) + 1);
    assert(!isBothDir(sm, v));
    periodicFunctions(sm, EVNT_DECISION, (int)(declev(sm->aq) + 1));
    return l;
}

void resetSchedImplied(satm_t *sm, uint v) {
	flexarray_t *fa;
	uint *a, i, sz;
    assert(v > 0 && v <= sm->varnum);
    //assert(sm->vt[v] != NULL);
    fa = sm->vt[v]->uip_implied;
    a = getFlexArray(fa);
    sz = getFlexArraySize(fa);
    for (i = 0; i < sz; i+=2) {
        uint x = getVar(a[i]);
        assert(x > 0 && x <= sm->varnum);
        decSchedCnt(sm, x);
    }
    resetFlexArray(fa);
}

void reset(satm_t *sm, const uint v) {
    var_t *vt = sm->vt[v];
    // must be a valid variable
    assert(v > 0 && v <= sm->varnum);
    // can't have MARKED or ADDED flags
    assert(!isMarked(sm, v) && !isAdded(sm, v));
    // only decision variables can have s/h-cubes
    assert(!hasCube(sm,v) || !getReason(sm, v));    
    // it can't be a variable eliminated in preprocessing
    //assert(sm->vt[v] != NULL);
    // if not assigned, can't have s/h-cube
    assert(!(!isAssigned(sm,v) && hasCube(sm,v)));
    // dependant vars can't have s/h-cube
    assert(!isDependant(sm,v) || !hasCube(sm,v));
    // dependant vars are always implied
    assert(!isDependant(sm,v) || !isBothDir(sm,v));
    // dependant vars must always have an antecedent clause
    assert(!isDependant(sm,v) || getReason(sm,v));
    // implied variables can't have scheduled implied vars
    assert(!getReason(sm,v) || 
            getFlexArraySize(vt->uip_implied) == 0);
    // if it's not assigned, must be implied
    assert(isAssigned(sm,v) || getReason(sm, v));
    // implied variables can't have any BCTs associated
    assert(!getReason(sm,v) || !vt->hc[0]->nd);
    assert(!getReason(sm,v) || !vt->hc[1]->nd);
    // implied vars can't be BCT literal or branch nodes
    assert(!getReason(sm,v) || !isLiteralNd(sm,v));
    assert(!getReason(sm,v) || !isBranchNd(sm,v));

    setDecLev(sm, v, 0);
    clearAssigned(sm, v);

    if (getReason(sm, v)) { // implied var
        unregReason(sm, v);    
        if (isDependant(sm, v)) {
            return;
        }
    } else if (isBothDir(sm, v)) {
        // decision var tried both ways
        resetSchedImplied(sm, v);
        clearCube(sm, v);
        sm->typs[v] &= (FIXED | FIXED_SIGN);
        assert(!vt->hc[0]->nd);
        assert(!vt->hc[1]->nd);
    } else if (!hasCube(sm, v)) {
        // decision vars tried one-way, which
        // don't have a s/h-cube. as such vars
        // won't be flipped, their queues of
        // scheduled vars need to be reset
        resetSchedImplied(sm, v);
#ifdef DEBUG
        clearBranchNd(sm,v);
#endif
        clearCube(sm, v);
        // if the variable was a branch node of a BCT
        // which wasn't involved in any conflicts, it
        // will be backtracked over. in that case, its
        // stored sub-BCT needs to be reset properly.
        assert(!vt->hc[0]->nd);
        assert(!vt->hc[1]->nd);
    }
    // the variable can't have a registered reason 
    // at this point
    assert(getReason(sm, v) == NULL);
    
    fh_node_insert(sm->fheap[vt->fheap_idx], vt->fnode, 
            -computeVarPriority(sm, v));
    sm->fheap_ptr = vt->fheap_idx;
}

inline uint isLitSat(satm_t *sm, uint lit) {
    uint v = getVar(lit);
    assert(isAssigned(sm, v));
    return getPhase(lit) == getSign(sm, v);
}

uint compute(satm_t *sm, const uint lit) {
    uint *usp;
    uint v = getVar(lit), ph = getPhase(lit);
    stack_t *todo = sm->cls[CLSIDX(v, (ph ^ 1))];
#ifdef __GNUC__
    uint st_sz = _umax(stackSize(todo), stackSize(sm->cls[CLSIDX(v,2)]));
    uint ustack[st_sz];
#else
    uint *ustack = sm->tstack;
#endif
    usp = ustack;
    appendAQ(sm->aq, lit, getDecLev(sm, v), !getReason(sm, v));
    markAssigned(sm, v);
    setSign(sm, v, ph);
    reverseStack(todo);

solve_cnf_cls:
    if (!stackEmpty(todo)) {
        do {
            uint cl = (uint)stackPop(todo);
            cla_t *c = (cla_t *)TRIM(cl);
            if (isAssigned(sm, getVar(c[LSB2(cl) ^ 3]))) {
                usp[0] = cl;
                usp++;
                continue;
            }
            {
                uint i = 3;
                uint csz = getClSize(c);
                while (i < csz) {
                    uint xl = c[i++];
                    uint xv = getVar(xl);
                    uint xph = getPhase(xl);
                    // find unassigned or assigned and satisfied literal
                    if (!isAssigned(sm, xv) || xph == getSign(sm, xv)) {
                        swap((int *)c, (int)(i - 1), (int)LSB2(cl));
                        //assert(sm->vt[xv] != NULL);
                        stackPush(sm->cls[CLSIDX(xv,xph)], (Item)cl);
                        goto solve_cnf_cls;
                    }    
                }
            }
            usp[0] = cl;
            usp++;
            {
                uint ulit = c[LSB2(cl) ^ 3];
                if (bcpqPut(sm, ulit, c) == BCP_CONFLICT) {
                    assert(getDecLev(sm, getVar(ulit)) == declev(sm->aq));
                    sm->conflicting_cl = c;
                    stackFillNC(todo, ustack, usp);
                    assert(bcpqEmpty(sm->impq));
                    return CONFLICT;
                }
            }
        } while (!stackEmpty(todo));
    }

    stackFillNC(todo, ustack, usp);
    todo = sm->cls[CLSIDX(v,2)];
    // this stack reversing is just a hack that improves the
    // performance slightly for unknown reasons...
    reverseStack(todo);

    // undo stack now contains clauses which need to be
    // searched for unassigned vars, while todo stack contains
    // clauses which definitely imply dependant variable
    usp = ustack;

inner_coe_loop:
    if (!stackEmpty(todo)) {
        do {
            uint cl = (uint)stackPop(todo);
            uint wlit, sgn = FALSE;
            cla_t *c = (cla_t *)TRIM(cl);
            wlit = c[LSB2(cl) ^ 3];
            if (isAssigned(sm, wlit)) {
                usp[0] = cl;
                usp++;
                continue;
            }
            {
                uint i = 3;
                uint csz = getClSize(c);
                do {
                    uint xv = c[i];
                    if (!isAssigned(sm, xv)) {
                        swap((int *)c, (int)i, (int)LSB2(cl));
                        //assert(sm->vt[xv] != NULL);
                        stackPush(sm->cls[CLSIDX(xv,2)], (Item)cl);
                        goto inner_coe_loop;
                    }
                    sgn ^= (getSign(sm, xv) ^ 1);
                } while (++i < csz);
            }
            usp[0] = cl;
            usp++;
            {
                assert(!isAssigned(sm, wlit));
                sgn ^= (ph ^ 1);
                sgn = (sgn != getEquivSgn(c));
                if (bcpqPut(sm, makeLit(wlit, sgn), c) == BCP_CONFLICT) {
                    sm->conflicting_cl = c;
                    stackFillNC(todo, ustack, usp);
                    return CONFLICT;
                }
            }
        } while (!stackEmpty(todo));
    }
    stackFillNC(todo, ustack, usp);
    return NO_CONFLICT;        
}

satm_t *newManager(uint varnum, uint clnum, flexarray_t *elimvars,
        flexarray_t *depvars, flexarray_t *clusters, stats_t *stats) {
	uint i, sz, fsz = 16, ssz = 8, *a;
    satm_t *sm = xcalloc(1, sizeof(satm_t) +
        (varnum + 1) * sizeof(var_t *));
    sm->typs = (uintsh *)xcalloc(1, (varnum + 1) * sizeof(uintsh));
    sm->cls = (stack_t **)xcalloc(1, 3 * 
            (varnum + 1) * sizeof(stack_t *));
    sm->status = NONE;
    sm->varnum = varnum;
    sm->clnum = clnum;
    sm->conflict = initFlexArray(fsz);
    sm->tmp_fa = initFlexArray(fsz);
    sm->progress = initFlexArray(fsz);
    sm->tmp_stack = initStack(ssz);
    sm->tounmark = initStack(ssz);
    sm->hcs = initHcSupport(varnum, (uint)sm, &getVarPriority);
    sm->hcs->follow_all_branches = TRUE;
    sm->hcs->treshold = (float)0.5;
    sm->hcs->max_bnodes = 20;
    sm->stats = stats;
    sm->impq = initBcpq(varnum);
#ifndef __GNUC__
    sm->tstack = xmalloc(sizeof(uint) * varnum);
#endif
    sm->aq = initAQ(varnum);
    //qsortFlexArray(elimv);
    sm->unit_cls = initStack(ssz);
    //assert(checkSortFlexArray(elimv));
    
    // initialize var_t structs
    for (i = 1; i <= varnum; i++) {
		var_t *vt = xcalloc(1, sizeof(var_t));
        sm->vt[i] = vt;
    }
    
    // set ELIMINATED & DEPENDANT flags
    sz = getFlexArraySize(elimvars);
    a = getFlexArray(elimvars);
    for (i = 0; i < sz; i++) {
        uint v = a[i];
        markEliminated(sm, v);
        //free(sm->vt[v]);
        //sm->vt[v] = NULL;
    }
    sz = getFlexArraySize(depvars);
    a = getFlexArray(depvars);
    for (i = 0; i < sz; i++) {
        markDependant(sm, a[i]);
    }

    // initialize fheap
    sm->fheap_ptr = 0;
    if (clusters && getFlexArraySize(clusters) > 1) {
        uint csz = getFlexArraySize(clusters);
        sm->fheap_sz = csz;
        a = getFlexArray(clusters);
        sm->fheap = (fheap_t **)xmalloc(sm->fheap_sz * sizeof(fheap_t *));
        for (i = 0; i < csz; i++) {
            uint *cl, j, clust_sz = getFlexArraySize((flexarray_t *)a[i]);
            cl = getFlexArray((flexarray_t *)a[i]);
            sm->fheap[i] = fh_alloc((int)sm->varnum + 1);
            for (j = 0; j < clust_sz; j++) {
                uint v = cl[j];
                if (v > sm->varnum || !v) {
                    warrn("Invalid variable in the cluster!");
                    continue;
                }
                if (isEliminated(sm, v)) continue;
                sm->vt[v]->fheap_idx = i;    
            } 
        }
    } else {
        sm->fheap_sz = 1;
        sm->fheap = (fheap_t **)xmalloc(sizeof(fheap_t *));
        sm->fheap[0] = fh_alloc((int)sm->varnum + 1);
    }

    // finish initialization
    for (i = 1; i <= varnum; i++) {
        if (!isEliminated(sm, i)) {
            // variable i is completely eliminated from the
            // CNF and should never appear anywhere, so its
            // var_t struct can be freed
            //free(sm->vt[i]);
            //sm->vt[i] = NULL;
        //} else {
            var_t *vt = sm->vt[i];
            uint j;
            //vt->fheap_idx = 0;
            vt->uip_implied = initFlexArray(4);
            for (j = 0; j < 2; j++) {
                vt->hc[j] = xcalloc(1, sizeof(hc_anchor_t));
            }
            vt->fnode = fh_create((int)i);
            for (j = 0; j < 3; j++) {
                sm->cls[CLSIDX(i,j)] = initStack(ssz);
            }
        }
    }

    // set pointers to heuristic functions
    sm->hfunc = hfunc1;
    sm->init_hfunc = hfunc2;

    return sm;
}

void initManager(satm_t *sm, flexarray_t *cnf, flexarray_t *coe,
        uint **pri_mtx, clock_t tout) {
    setParams(sm, NULL);
    sm->stats->cacheSize = sm->params->cacheSize;
    sm->stats->cacheDelPerc = sm->params->cacheDelPerc;
    sm->stats->cacheInflationRate = sm->params->cacheInflationRate;
    sm->learn1UIP = TRUE;
    sm->timeout = (tout) ? tout : (clock_t)-1;
    init_lit_priorities(sm, pri_mtx);
    add_CNF_lst(sm, cnf);
    add_COE_lst(sm, coe);
    sm->learned_cls = xmalloc(sm->params->cacheSize * sizeof(uint));
    sm->lcls_ptr = (uint)0;
}

void freeDeadClause(satm_t *sm, cla_t *cl) {
    int i;
    assert(cl != NULL);
    assert(isLearned(cl));
    assert(!isUnitCl(cl));
    for (i = 1; i < 3; i++) {
        int k;
        stack_t *w;
        uint l = cl[i];
        uint v = getVar(l);
        //assert(sm->vt[v] != NULL);
        assert(v > 0 && v <= sm->varnum);
        w = sm->cls[CLSIDX(v,getPhase(l))];
        k = findInStack(w, (Item)((uint)cl | i));
        assert(k >= 0);
        stackDelEl(w, (uint)k); 
    }
    decNumLearned(sm);
    free(cl);
}

void freeManager(satm_t *sm) {
    uint i, j;
    for (i = 1; i <= sm->varnum; i++) {
        var_t *v = sm->vt[i];
        if (isEliminated(sm, i)) continue;
        free(v->fnode);
        v->fnode = NULL;
        freeFlexArray(v->uip_implied);
        v->uip_implied = NULL;
        unregReason(sm, i);
    }
    for (i = 0; i < sm->lcls_ptr; i++) {
		cla_t *c;
        assert(sm->learned_cls[i]);
        c = (cla_t *)sm->learned_cls[i];
        assert(c != NULL);
        decNumLearned(sm);
        free(c);    
    }
    free(sm->learned_cls);
    sm->learned_cls = NULL;
    while (!stackEmpty(sm->unit_cls)) {
        cla_t *cl = (cla_t *)stackPop(sm->unit_cls);
        decNumLearned(sm);
        free(cl);
    }
    
    assert(sm->stats->clNumLearned == 0);

    for (i = 1; i <= sm->varnum; i++) {
        var_t *v = sm->vt[i];
        if (isEliminated(sm, i)) continue;
        for (j = 0; j < 3; j++) {
            stackFree(sm->cls[CLSIDX(i,j)]);
        }
        freeHcAnchor(sm->hcs, v->hc[0]);
        free(v->hc[1]);
        free(v->fnode);
        free(v);
        sm->vt[i] = NULL;
    }
    free(sm->typs);
    sm->typs = NULL;
    for (i = 0; i < sm->fheap_sz; i++) {
        fh_free(sm->fheap[i]);
        sm->fheap[i] = NULL;
    }
    free(sm->fheap);
    sm->fheap = NULL;
    freeAQ(sm->aq);
    sm->aq = NULL;
    freeBcpq(sm->impq);
    sm->impq = NULL;
    freeHcSupport(sm->hcs);
    sm->hcs = NULL;
    free(sm->params);
    sm->params = NULL;
    stackFree(sm->tounmark);
    sm->tounmark = NULL;
    stackFree(sm->tmp_stack);
    sm->tmp_stack = NULL;
    freeFlexArray(sm->tmp_fa);
    sm->tmp_fa = NULL;
    freeFlexArray(sm->progress);
    sm->progress = NULL;
    freeFlexArray(sm->conflict);
    sm->conflict = NULL;
    stackFree(sm->unit_cls);
    sm->unit_cls = NULL;
#ifndef __GNUC__
    free(sm->tstack);
#endif
    free(sm); 
}

void setParams(satm_t *sm, params_t *params) {
    params_t *p = params;
    if (!p) {
        p = xcalloc(1, sizeof(params_t));
        // set default values
        p->priInitial = PRI_SIMPLIFIED;
        p->impLongChains = TRUE;
        p->priUpdatePolicy = PRI_UIP_CONSTRUCTION;
        p->priIncrement = 2;
        p->priScalePeriod = 300;
        p->priBinBoost = 2; 
        p->priLitsUpdated = 4;
        p->cacheUpdatePolicy = UIP_CONFLICTING_CLS;
        p->cacheSize = (uint)1 << 14; // 16384
        p->maxCacheSize = (uint)1 << 19; // 524288
        p->cacheDelPerc = (uint)50;
        p->cacheInflationRate = 1.5;
        p->faResizeTresh = 64;
    }    
    sm->params = p;
}

/*
// find the length of the longest path in BCT nd
uint findDeepest(hc_node_t *nd, stack_t *s) {
    uint deepest = 0, current = 0;
    hc_node_t *x = nd;
    assert(stackEmpty(s));
    while (x) {
        if (x == TERM_ND) {
            if (current > deepest) {
                deepest = current;
            }
            if (!stackEmpty(s)) {
                current = (uint)stackPop(s);
                x = (hc_node_t *)stackPop(s);
            } else {
                break;
            }
        } else if (hc_is_bnode(x)) {
            current++;
            stackPush(s, (Item)x->nxt);
            stackPush(s, (Item)current);
            x = x->neg;
        } else {
            current++;
            x = x->nxt;
        }
    }
    assert(stackEmpty(s));
    return deepest;
}// */

uint chooseBranch(satm_t *sm, hc_node_t *bnd) {
    uint sub = subsumption(sm->hcs, bnd->neg, bnd->nxt);
    if (sub == NONE) {
        uint bndv, posb, negb;
        assert(bnd->neg != TERM_ND);
        assert(bnd->nxt != TERM_ND);
        bndv = getVar(bnd->var);
        markCube(sm, bndv);
        posb = hc_is_bnode(bnd->nxt);
        negb = hc_is_bnode(bnd->neg);
        if (posb && negb) { // both are branch nodes
            var_t *vt = sm->vt[bndv];
            return vt->pri[TRUE] > vt->pri[FALSE];
        } else if (posb) { // neg is literal node
            return FALSE;
        } else if (negb) { // nxt (pos) is literal node
            return TRUE;
        } else { // both are literal nodes
            uint posv, negv;
            posv = getVar(bnd->nxt->var);
            negv = getVar(bnd->neg->var);
            return computeVarPriority(sm, posv) < 
                computeVarPriority(sm, negv);
        }
    } else {
        return sub;
    }
}

#define CLOCK_MASK 0xFFF

void search(satm_t *sm) {
    uint nxt = 0;
    uint v = 0;
    uint clk = 0;

    // sanity checks
    assert(sm->num_marked == 0);
    assert(getFlexArraySize(sm->conflict) == 0);
    assert(getFlexArraySize(sm->tmp_fa) == 0);
    assert(sm->conflicting_cl == NULL);
    assert(stackEmpty(sm->tmp_stack));
    assert(stackEmpty(sm->tounmark));
    assert(bcpqEmpty(sm->impq));

    do {
        // WARRNING: clock_t will generally be int on linux
        // and uint on cygwin. if not set, timeout will be
        // -1, so a cast is needed to make it work on both OSes.
        if (!(clk++ & CLOCK_MASK) && 
                ((ulong)clock() >= (ulong)sm->timeout)) {
            return;
        }
        if (numScheduledAQ(sm->aq) == 0) { // make a new decision
            nxt = decide(sm);
            if (nxt == 0) {
                setManagerStatus(sm, TRUE);
                return;
            } else {
#ifdef DEBUG
                var_t *vt;
#endif
                v = getVar(nxt);
#ifdef DEBUG
                vt = sm->vt[v];
                assert(!vt->hc[0]->nd);
                assert(!vt->hc[1]->nd);
#endif
            }
        } else {
            uint data;
            uint status = getScheduledAQ(sm->aq, &nxt, &data);
            hc_node_t *bnd, *nd;
            v = getVar(nxt);
            assert(v > 0 && v <= sm->varnum);
            switch (status) {
                case SCHD_DECISION:
                    break;
                case SCHD_HC_STEM:
                    assert(!hasCube(sm, v));
                    if (isAssigned(sm, v)) {
                        if (isLitSat(sm, nxt)) {
                            continue;
                        } else {
                            periodicFunctions(sm, EVNT_USED_HCUBE_NODE, 0);
                            sm->conflicting_cl = getReason(sm, v);
                            assert(sm->conflicting_cl != NULL);
                            // don't learn new UIP clause, as the clause
                            // can be incorrect
                            sm->learn1UIP = FALSE;
                            bcpqReset(sm->impq);
                            // append conflicting variable to the
                            // conflict -> this is required by B-cubing
                            // theoretical framework (makes the proof
                            // easier)
                            appendToFlexArray(sm->conflict, v);
                            addVar(sm, v);
                            goto resolve_conflict;
                        }
                    } else {
                        periodicFunctions(sm, EVNT_USED_HCUBE_NODE, 0);
                        markBothDir(sm, v);
#ifdef DEBUG
                        markLiteralNd(sm, v);
#endif
                        setDecLev(sm, v, declev(sm->aq) + 1);
                    }
                    break;
                case SCHD_HC_BRANCH:
                    assert(!hasCube(sm, v));
                    assert(numScheduledAQ(sm->aq) == 0);
                    bnd = (hc_node_t *)data;
                    assert((uint)bnd->nxt > 1 || (uint)bnd->neg > 1);
                    periodicFunctions(sm, EVNT_USED_HCUBE_BRANCH, 0);
                    if (isAssigned(sm, v)) {
                        uint ph = getSign(sm, v);
                        if (ph) {
                            freeHcNode(sm->hcs, bnd->neg);
                            bnd->nxt = TRIM(bnd->nxt);
                            bnd->neg = NULL;
                        } else {
                            freeHcNode(sm->hcs, bnd->nxt);
                            bnd->nxt = TRIM(bnd->neg);
                            bnd->neg = NULL;
                        }
                        scheduleHcube(sm, bnd->nxt);
                        // hcs counters have been changed when the
                        // other (not visited) branch was freed, so
                        // the counters need to be reset to zeros.
                        resetHcSupport(sm->hcs);
                        continue;
                    } else {
                        var_t *vt = sm->vt[v];
                        // choose the BCT branch to follow
                        data = chooseBranch(sm, bnd); 
                        nd = (data) ? bnd->nxt : bnd->neg;
                        // schedule all literals in the branch up to
                        // the next branch node
                        scheduleHcube(sm, nd);
#ifdef DEBUG
                        markBranchNd(sm, v);
#endif
                        nxt = makeLit(v, data);
                        setDecLev(sm, v, declev(sm->aq) + 1);
                        assert(bnd != NULL && bnd != TERM_ND);
                        // register the branch
                        vt->hc[1]->nd = bnd;
                    }
                    break;
            }
        }

        assert(nxt);
        assert(v > 0 && v <= sm->varnum);
        assert(!isAssigned(sm, v));

        if (compute(sm, nxt) == CONFLICT) {
            goto resolve_conflict;
        }

        while (!bcpqEmpty(sm->impq)) {
            periodicFunctions(sm, EVNT_IMPLICATION, 0);
            nxt = bcpqGet(sm);
            v = getVar(nxt);
            assert(v > 0 && v <= sm->varnum);
            assert(!isAssigned(sm, v));
            setDecLev(sm, v, declev(sm->aq));
            if (compute(sm, nxt) == CONFLICT) {
                goto resolve_conflict;
            }
        }
        continue;

resolve_conflict:
        if (resolve(sm) == CONFLICT) {
            setManagerStatus(sm, FALSE);
            return;
        }

    } while (TRUE);
}

void setManagerStatus(satm_t *sm, char status) {
    sm->status = (uint)status;
}

#ifdef DEBUG
inline char checkClSanity(satm_t *sm, cla_t *c) {
    uint i, sz;
    assert(c != NULL);
    assert(!isTautology(c));
    sz = getClSize(c);

    if (isEquiv(c)) {
        assert(sz > 3);
        for (i = 1; i < sz; i++) {
            uint v = c[i];
            assert(v > 0 && v <= sm->varnum);
        }
        assert(isDependant(sm, c[sz]));
    } else {
        for (i = 1; i < sz; i++) {
            uint v = getVar(c[i]);
            assert(v > 0 && v <= sm->varnum);
        }
    }
    return TRUE;
}
#endif

/*
 * Revision history:
 *
 * $Log: hsat.h,v $
 * Revision 1.13  2005/02/21 23:44:09  root
 * Minor update.
 *
 * Revision 1.12  2005/02/21 22:32:30  root
 * SAT competition release.
 *
 * Revision 1.11  2005/02/14 23:14:18  root
 * Cleaned up the code a bit and did some minor optimizations.
 *
 * Revision 1.10  2005/02/09 01:57:08  root
 * B-cubing adapted to B-cubing theory.
 *
 * Revision 1.9  2005/01/30 13:26:18  root
 * Fixed bug in possible skipping over branch BCT nodes.
 *
 * Revision 1.8  2004/12/23 23:07:34  root
 * Optimized B-cubing implementation (now the algorithm picks the
 * variable with the highest priority as the new branch node and
 * search backtracks over unused BCT branch nodes). Fixed some
 * minor bugs. Code passes splint -weak check.
 *
 * Revision 1.6  2004/12/13 07:44:51  root
 * Adapted the code to VC C++ .NET compiler and 
 * made some minor optimizations.
 *
 * Revision 1.5  2004/12/11 04:46:12  root
 * Enofrced stricter c89 standard conformance.
 *
 * Revision 1.4  2004/12/09 18:52:41  root
 * Added B-Cubing (originally called Hypercubing).
 *
 * Revision 1.3  2004/10/31 19:10:56  root
 * Minor modifications.
 *
 * Revision 1.2  2004/10/24 00:06:46  root
 * Fixed bug in counting scheduled implied vars. 
 * Added a few assertions.
 *
 * Revision 1.1  2004/10/12 23:31:15  root
 * Initial import.
 *
 *
 */
